@@ -5,8 +5,7 @@
 //! # #[async_std::main]
 //! # async fn main() -> Result<(), Box<dyn std::error::Error>> {
 //! # use ipfs_embed::{Config, DefaultParams, Ipfs};
-//! # let cache_size = 100;
-//! let ipfs = Ipfs::<DefaultParams>::new(Config::new(None, cache_size)).await?;
+//! let ipfs = Ipfs::<DefaultParams>::new(Config::default()).await?;
 //! ipfs.listen_on("/ip4/0.0.0.0/tcp/0".parse()?)?;
 //! # Ok(()) }
 //! ```
@@ -14,14 +13,16 @@ use crate::db::StorageService;
 pub use crate::db::{StorageConfig, TempPin};
 pub use crate::net::{
     generate_keypair, AddressRecord, AddressSource, BitswapConfig, BroadcastConfig, DnsConfig,
-    Event, GossipsubConfig, IdentifyConfig, KadConfig, Key, Keypair, ListenerEvent, ListenerId,
-    MdnsConfig, Multiaddr, NetworkConfig, PeerId, PeerInfo, PeerRecord, PingConfig, Quorum, Record,
-    SyncEvent, SyncQuery, ToLibp2p, TransportConfig,
+    DocId, Event, GossipEvent, GossipsubConfig, Head, IdentifyConfig, KadConfig, Key, Keypair,
+    ListenerEvent, ListenerId, LocalStreamWriter, MdnsConfig, Multiaddr, NetworkConfig, PeerId,
+    PeerInfo, PeerRecord, PingConfig, PublicKey, Quorum, Record, SecretKey, SignedHead, StreamId,
+    StreamReader, SwarmEvents, SyncEvent, SyncQuery, ToLibp2p, TransportConfig,
 };
 use crate::net::{BitswapStore, NetworkService};
 #[cfg(feature = "telemetry")]
 pub use crate::telemetry::telemetry;
 use async_trait::async_trait;
+pub use db::Batch;
 use executor::Executor;
 use futures::stream::Stream;
 use libipld::codec::References;
@@ -34,6 +35,7 @@ use libp2p::kad::kbucket::Key as BucketKey;
 pub use libp2p::multiaddr;
 use prometheus::Registry;
 use std::collections::HashSet;
+use std::path::Path;
 use std::sync::Arc;
 
 mod db;
@@ -56,14 +58,20 @@ pub struct Config {
 impl Config {
     /// Creates a default configuration from a `path` and a `cache_size`. If the `path` is `None`,
     /// ipfs will use an in-memory block store.
-    pub fn new(path: Option<std::path::PathBuf>, cache_size: u64) -> Self {
+    pub fn new(path: &Path, keypair: Keypair) -> Self {
         let sweep_interval = std::time::Duration::from_millis(10000);
-        let storage = StorageConfig::new(path, cache_size, sweep_interval);
-        let network = NetworkConfig::default();
+        let storage = StorageConfig::new(Some(path.join("blocks")), 0, sweep_interval);
+        let network = NetworkConfig::new(path.join("streams"), keypair);
         Self { storage, network }
     }
 }
 
+impl Default for Config {
+    fn default() -> Self {
+        Self::new(Path::new("."), generate_keypair())
+    }
+}
+
 /// Ipfs node.
 #[derive(Clone)]
 pub struct Ipfs<P: StoreParams> {
@@ -121,6 +129,11 @@ where
         Ok(Self { storage, network })
     }
 
+    /// Returns the local `PublicKey`.
+    pub fn local_public_key(&self) -> PublicKey {
+        self.network.local_public_key()
+    }
+
     /// Returns the local `PeerId`.
     pub fn local_peer_id(&self) -> PeerId {
         self.network.local_peer_id()
@@ -210,6 +223,11 @@ where
         Ok(())
     }
 
+    /// Returns true if the dht was bootstrapped.
+    pub fn is_bootstrapped(&self) -> bool {
+        self.network.is_bootstrapped()
+    }
+
     /// Gets the closest peer to a key. Useful for finding the `Multiaddr` of a `PeerId`.
     pub async fn get_closest_peers<K>(&self, key: K) -> Result<()>
     where
@@ -251,7 +269,7 @@ where
 
     /// Subscribes to a `topic` returning a `Stream` of messages. If all `Stream`s for
     /// a topic are dropped it unsubscribes from the `topic`.
-    pub fn subscribe(&self, topic: &str) -> Result<impl Stream<Item = Vec<u8>>> {
+    pub fn subscribe(&self, topic: &str) -> Result<impl Stream<Item = GossipEvent>> {
         self.network.subscribe(topic)
     }
 
@@ -353,6 +371,22 @@ where
         self.storage.flush().await
     }
 
+    /// Perform a set of storage operations in a batch, and discards the result.
+    ///
+    /// You should keep a batch open for as short as possible.
+    /// You *must not* create a batch inside a batch, or use the outer ipfs.
+    pub fn read_batch<F: FnOnce(&mut Batch<'_, P>) -> Result<R>, R>(&self, f: F) -> Result<R> {
+        self.storage.ro("read_batch", f)
+    }
+
+    /// Perform a set of storage operations in a batch, and stores the result.
+    ///
+    /// You should keep a batch open for as short as possible.
+    /// You *must not* create a batch inside a batch, or use the outer ipfs.
+    pub fn write_batch<R>(&self, f: impl FnOnce(&mut Batch<'_, P>) -> Result<R>) -> Result<R> {
+        self.storage.rw("write_batch", f)
+    }
+
     /// Registers prometheus metrics in a registry.
     pub fn register_metrics(&self, registry: &Registry) -> Result<()> {
         self.storage.register_metrics(registry)?;
@@ -361,8 +395,58 @@ where
     }
 
     /// Subscribes to the swarm event stream.
-    pub fn swarm_events(&self) -> impl Stream<Item = Event> {
-        self.network.event_stream()
+    pub fn swarm_events(&self) -> SwarmEvents {
+        self.network.swarm_events()
+    }
+
+    /// Returns the documents both local and replicated.
+    pub fn docs(&self) -> Result<Vec<DocId>> {
+        self.network.docs()
+    }
+
+    /// Returns the streams both local and replicated.
+    pub fn streams(&self) -> Result<Vec<StreamId>> {
+        self.network.streams()
+    }
+
+    /// Returns the streams both local and replicated for the given document id.
+    pub fn substreams(&self, doc: DocId) -> Result<Vec<StreamId>> {
+        self.network.substreams(doc)
+    }
+
+    /// Adds the peers to a replicated stream.
+    pub fn stream_add_peers(&self, doc: DocId, peers: impl Iterator<Item = PeerId>) {
+        self.network.stream_add_peers(doc, peers)
+    }
+
+    /// Returns the current head of a stream.
+    pub fn stream_head(&self, id: &StreamId) -> Result<Option<SignedHead>> {
+        self.network.stream_head(id)
+    }
+
+    /// Returns a reader for the slice.
+    pub fn stream_slice(&self, id: &StreamId, start: u64, len: u64) -> Result<StreamReader> {
+        self.network.stream_slice(id, start, len)
+    }
+
+    /// Removes a local or replicated stream.
+    pub fn stream_remove(&self, id: &StreamId) -> Result<()> {
+        self.network.stream_remove(id)
+    }
+
+    /// Returns a writter to append to a local stream.
+    pub fn stream_append(&self, id: DocId) -> Result<LocalStreamWriter> {
+        self.network.stream_append(id)
+    }
+
+    /// Subscribes to a replicated stream.
+    pub fn stream_subscribe(&self, id: &StreamId) -> Result<()> {
+        self.network.stream_subscribe(id)
+    }
+
+    /// Updates the head of a replicated stream.
+    pub fn stream_update_head(&self, head: SignedHead) {
+        self.network.stream_update_head(head)
     }
 }
 
@@ -423,6 +507,7 @@ where
 #[cfg(test)]
 mod tests {
     use super::*;
+    use async_std::future::timeout;
     use futures::join;
     use futures::stream::StreamExt;
     use libipld::cbor::DagCborCodec;
@@ -431,6 +516,7 @@ mod tests {
     use libipld::store::DefaultParams;
     use libipld::{alias, ipld};
     use std::time::Duration;
+    use tempdir::TempDir;
 
     fn tracing_try_init() {
         tracing_subscriber::fmt()
@@ -439,11 +525,12 @@ mod tests {
             .ok();
     }
 
-    async fn create_store(enable_mdns: bool) -> Result<Ipfs<DefaultParams>> {
+    async fn create_store(enable_mdns: bool) -> Result<(Ipfs<DefaultParams>, TempDir)> {
+        let tmp = TempDir::new("ipfs-embed")?;
         let sweep_interval = Duration::from_millis(10000);
         let storage = StorageConfig::new(None, 10, sweep_interval);
 
-        let mut network = NetworkConfig::default();
+        let mut network = NetworkConfig::new(tmp.path().into(), generate_keypair());
         if !enable_mdns {
             network.mdns = None;
         }
@@ -453,7 +540,7 @@ mod tests {
             .next()
             .await
             .unwrap();
-        Ok(ipfs)
+        Ok((ipfs, tmp))
     }
 
     fn create_block(bytes: &[u8]) -> Result<Block<DefaultParams>> {
@@ -463,7 +550,7 @@ mod tests {
     #[async_std::test]
     async fn test_local_store() -> Result<()> {
         tracing_try_init();
-        let store = create_store(false).await?;
+        let (store, _tmp) = create_store(false).await?;
         let block = create_block(b"test_local_store")?;
         let tmp = store.create_temp_pin()?;
         store.temp_pin(&tmp, block.cid())?;
@@ -477,8 +564,8 @@ mod tests {
     #[ignore] // test is too unreliable for ci
     async fn test_exchange_mdns() -> Result<()> {
         tracing_try_init();
-        let store1 = create_store(true).await?;
-        let store2 = create_store(true).await?;
+        let (store1, _tmp) = create_store(true).await?;
+        let (store2, _tmp) = create_store(true).await?;
         let block = create_block(b"test_exchange_mdns")?;
         let tmp1 = store1.create_temp_pin()?;
         store1.temp_pin(&tmp1, block.cid())?;
@@ -497,9 +584,9 @@ mod tests {
     #[ignore] // test is too unreliable for ci
     async fn test_exchange_kad() -> Result<()> {
         tracing_try_init();
-        let store = create_store(false).await?;
-        let store1 = create_store(false).await?;
-        let store2 = create_store(false).await?;
+        let (store, _tmp) = create_store(false).await?;
+        let (store1, _tmp) = create_store(false).await?;
+        let (store2, _tmp) = create_store(false).await?;
 
         let addr = store.listeners()[0].clone();
         let peer_id = store.local_peer_id();
@@ -532,7 +619,7 @@ mod tests {
     #[async_std::test]
     async fn test_provider_not_found() -> Result<()> {
         tracing_try_init();
-        let store1 = create_store(true).await?;
+        let (store1, _tmp) = create_store(true).await?;
         let block = create_block(b"test_provider_not_found")?;
         if store1
             .fetch(block.cid(), vec![store1.local_peer_id()])
@@ -577,8 +664,8 @@ mod tests {
     #[async_std::test]
     async fn test_sync() -> Result<()> {
         tracing_try_init();
-        let local1 = create_store(false).await?;
-        let local2 = create_store(false).await?;
+        let (local1, _tmp) = create_store(false).await?;
+        let (local2, _tmp) = create_store(false).await?;
         local1.add_address(&local2.local_peer_id(), local2.listeners()[0].clone());
         local2.add_address(&local1.local_peer_id(), local1.listeners()[0].clone());
 
@@ -649,20 +736,29 @@ mod tests {
         let stores = [create_store(false).await?, create_store(false).await?];
         async_std::task::sleep(Duration::from_millis(100)).await;
         stores[0]
-            .bootstrap(&[(stores[1].local_peer_id(), stores[1].listeners()[0].clone())])
+            .0
+            .bootstrap(&[(
+                stores[1].0.local_peer_id(),
+                stores[1].0.listeners()[0].clone(),
+            )])
             .await?;
         stores[1]
-            .bootstrap(&[(stores[0].local_peer_id(), stores[0].listeners()[0].clone())])
+            .0
+            .bootstrap(&[(
+                stores[0].0.local_peer_id(),
+                stores[0].0.listeners()[0].clone(),
+            )])
             .await?;
         let key: Key = b"key".to_vec().into();
 
         stores[0]
+            .0
             .put_record(
                 Record::new(key.clone(), b"hello world".to_vec()),
                 Quorum::One,
             )
             .await?;
-        let records = stores[1].get_record(&key, Quorum::One).await?;
+        let records = stores[1].0.get_record(&key, Quorum::One).await?;
         assert_eq!(records.len(), 1);
         Ok(())
     }
@@ -681,33 +777,149 @@ mod tests {
         ];
         let mut subscriptions = vec![];
         let topic = "topic";
-        for store in &stores {
-            for other in &stores {
+        for (store, _) in &stores {
+            for (other, _) in &stores {
                 if store.local_peer_id() != other.local_peer_id() {
                     store.dial_address(&other.local_peer_id(), other.listeners()[0].clone());
                 }
             }
-            subscriptions.push(store.subscribe(topic)?);
         }
 
         async_std::task::sleep(Duration::from_millis(500)).await;
-
-        stores[0].publish(&topic, b"hello gossip".to_vec()).unwrap();
-
-        for subscription in &mut subscriptions[1..] {
-            let msg = subscription.next().await.unwrap();
-            assert_eq!(msg.as_slice(), &b"hello gossip"[..]);
+        // Make sure everyone is peered before subscribing
+        for (store, _) in &stores {
+            subscriptions.push(store.subscribe(topic)?);
         }
+        async_std::task::sleep(Duration::from_millis(500)).await;
 
         stores[0]
+            .0
+            .publish(&topic, b"hello gossip".to_vec())
+            .unwrap();
+
+        for (idx, subscription) in subscriptions.iter_mut().enumerate() {
+            let mut expected = stores
+                .iter()
+                .enumerate()
+                .filter_map(|(i, s)| {
+                    if i == idx {
+                        None
+                    } else {
+                        Some(s.0.local_peer_id())
+                    }
+                })
+                .flat_map(|p| {
+                    // once for gossipsub, once for broadcast
+                    vec![GossipEvent::Subscribed(p), GossipEvent::Subscribed(p)].into_iter()
+                })
+                .chain(if idx != 0 {
+                    // store 0 is the sender
+                    Box::new(std::iter::once(GossipEvent::Message(
+                        stores[0].0.local_peer_id(),
+                        b"hello gossip".to_vec().into(),
+                    ))) as Box<dyn Iterator<Item = GossipEvent>>
+                } else {
+                    Box::new(std::iter::empty())
+                })
+                .collect::<Vec<GossipEvent>>();
+            while !expected.is_empty() {
+                let ev = timeout(Duration::from_millis(100), subscription.next())
+                    .await
+                    .unwrap()
+                    .unwrap();
+                assert!(expected.contains(&ev));
+                if let Some(idx) = expected.iter().position(|e| e == &ev) {
+                    // Can't retain, as there might be multiple messages
+                    expected.remove(idx);
+                }
+            }
+        }
+
+        // Check broadcast subscription
+        stores[0]
+            .0
             .broadcast(&topic, b"hello broadcast".to_vec())
             .unwrap();
 
         for subscription in &mut subscriptions[1..] {
-            let msg = subscription.next().await.unwrap();
-            assert_eq!(msg.as_slice(), &b"hello broadcast"[..]);
+            if let GossipEvent::Message(p, data) = subscription.next().await.unwrap() {
+                assert_eq!(p, stores[0].0.local_peer_id());
+                assert_eq!(data[..], b"hello broadcast"[..]);
+            } else {
+                panic!()
+            }
         }
 
+        // trigger cleanup
+        stores[0]
+            .0
+            .broadcast(&topic, b"r u still listening?".to_vec())
+            .unwrap();
+
+        let mut last_sub = subscriptions.drain(..1).next().unwrap();
+        drop(subscriptions);
+        let mut expected = stores[1..]
+            .iter()
+            .map(|s| s.0.local_peer_id())
+            .flat_map(|p| {
+                // once for gossipsub, once for broadcast
+                vec![GossipEvent::Unsubscribed(p), GossipEvent::Unsubscribed(p)].into_iter()
+            })
+            .collect::<Vec<_>>();
+        while !expected.is_empty() {
+            let ev = timeout(Duration::from_millis(100), last_sub.next())
+                .await
+                .unwrap()
+                .unwrap();
+            assert!(expected.contains(&ev));
+            if let Some(idx) = expected.iter().position(|e| e == &ev) {
+                // Can't retain, as there might be multiple messages
+                expected.remove(idx);
+            }
+        }
+        Ok(())
+    }
+
+    #[async_std::test]
+    async fn test_batch_read() -> Result<()> {
+        tracing_try_init();
+        let tmp = TempDir::new("ipfs-embed")?;
+        let network = NetworkConfig::new(tmp.path().into(), generate_keypair());
+        let storage = StorageConfig::new(None, 1000000, Duration::from_secs(3600));
+        let ipfs = Ipfs::<DefaultParams>::new(Config { storage, network }).await?;
+        let a = create_block(b"a")?;
+        let b = create_block(b"b")?;
+        ipfs.insert(&a)?;
+        ipfs.insert(&b)?;
+        let has_blocks =
+            ipfs.read_batch(|db| Ok(db.contains(a.cid())? && db.contains(b.cid())?))?;
+        assert!(has_blocks);
+        Ok(())
+    }
+
+    #[async_std::test]
+    async fn test_batch_write() -> Result<()> {
+        tracing_try_init();
+        let tmp = TempDir::new("ipfs-embed")?;
+        let network = NetworkConfig::new(tmp.path().into(), generate_keypair());
+        let storage = StorageConfig::new(None, 1000000, Duration::from_secs(3600));
+        let ipfs = Ipfs::<DefaultParams>::new(Config { storage, network }).await?;
+        let a = create_block(b"a")?;
+        let b = create_block(b"b")?;
+        let c = create_block(b"c")?;
+        let d = create_block(b"d")?;
+        ipfs.write_batch(|db| {
+            db.insert(&a)?;
+            db.insert(&b)?;
+            Ok(())
+        })?;
+        assert!(ipfs.contains(a.cid())? && ipfs.contains(b.cid())?);
+        let _: anyhow::Result<()> = ipfs.write_batch(|db| {
+            db.insert(&c)?;
+            db.insert(&d)?;
+            anyhow::bail!("nope!");
+        });
+        assert!(!ipfs.contains(c.cid())? && ipfs.contains(b.cid())?);
         Ok(())
     }
 
@@ -716,8 +928,8 @@ mod tests {
     async fn test_bitswap_sync_chain() -> Result<()> {
         use std::time::Instant;
         tracing_try_init();
-        let a = create_store(true).await?;
-        let b = create_store(true).await?;
+        let (a, _tmp) = create_store(true).await?;
+        let (b, _tmp) = create_store(true).await?;
         let root = alias!(root);
 
         let (cid, blocks) = test_util::build_tree(1, 1000)?;
@@ -756,8 +968,8 @@ mod tests {
     async fn test_bitswap_sync_tree() -> Result<()> {
         use std::time::Instant;
         tracing_try_init();
-        let a = create_store(true).await?;
-        let b = create_store(true).await?;
+        let (a, _tmp) = create_store(true).await?;
+        let (b, _tmp) = create_store(true).await?;
         let root = alias!(root);
 
         let (cid, blocks) = test_util::build_tree(10, 4)?;
